"""Functions for constructing a KB entity."""
# Standard Modules
from typing import Dict, List

# Local Modules
from app.kb.entity.types import Alias, Aliases


def aliases(
    aliases: List[Alias],  # pylint: disable=redefined-outer-name
) -> Aliases:
    """Create aliases.

    Args:
        aliases: Language, value pairs.

    Returns:
        Aliases in KB structure.
    """
    acc: Dict = {}

    for alias in aliases:
        lang = alias["language"]

        if alias["value"]:  # Don't propagate aliases with falsy values.
            acc[lang] = [*acc.get(lang, []), alias]

    return acc



"""Part class."""
# Standard Modules
from dataclasses import dataclass
from typing import Dict, List, Optional

# Local Modules
from cofactr.schema.types import Document


@dataclass
class Part:  # pylint: disable=too-many-instance-attributes
    """Part."""

    id: str

    classification: List[str]
    description: Optional[str]
    documents: List[Document]
    hero_image: Optional[str]
    mpn: Optional[str]
    alt_mpns: List[str]
    mfr: Optional[str]  # manufacturer name.
    msl: Optional[str]
    package: Optional[str]
    specs: Dict[str, str]
    terminations: Optional[int]

    inventory_level: Optional[int]
    buyable: Optional[int]
    quotable: Optional[int]
    maybe: Optional[int]
    updated_at: Optional[str]



"""Helper functions."""
# Standard Modules
from functools import reduce
from operator import getitem
from typing import List

# Local Modules
from cofactr.kb.entity.types import Mainsnak


def get_path(data, keys, default=None):
    """Access a nested dictionary."""
    try:
        return reduce(getitem, keys, data)
    except (KeyError, IndexError):
        return default


def find_preferred(data: List[Mainsnak], default=None):
    """Find preferred value."""
    return next(
        (x for x in data if x.get("rank") == "preferred"), data[0] if data else default
    )


def drop_deprecated(data: List[Mainsnak]):
    """Drop deprecated values."""
    return filter(lambda x: x.get("rank") == "deprecated", data)


identity = lambda x: x



"""Part object."""
# Local Modules
from cofactr.kb.entity.types import DataValue, Mainsnak
from cofactr.helpers import identity, get_path

TYPE_TO_RENDERER = {
    "monolingual_text": lambda v: v["text"],
    "quantity": lambda v: v["amount"],
}


def datavalue_to_str(datavalue: DataValue) -> str:
    """Render data-value to string.

    Example in:
        ```
        {
            "value": "CC0603JRNPOABN100",
            "type": "string"
        }
        ```

    Example out:
        `"CC0603JRNPOABN100"`
    """
    renderer = TYPE_TO_RENDERER.get(datavalue["type"], identity)

    return str(renderer(datavalue["value"]))


def mainsnak_to_str(mainsnak: Mainsnak) -> str:
    """Render mainSNAK to string."""
    return datavalue_to_str(get_path(mainsnak, keys=["mainsnak", "datavalue"]))



"""Part seller class."""
# Standard Modules
from dataclasses import dataclass
from typing import List, Optional


@dataclass
class Seller:  # pylint: disable=too-many-instance-attributes
    """Part seller."""

    id: str
    label: str
    aliases: List[str]
    authenticity_score: Optional[int]
    availability_score: Optional[int]
    additional_markup: Optional[float]
    additional_fee: Optional[float]
    certifications: List[str]
    #  A separate flag that's independent from accuracy.
    is_buyable: bool
    lead: Optional[int]
    preferred_distributor_group: Optional[str]



"""Cofactr graph API client."""
# Python Modules
from concurrent.futures import ThreadPoolExecutor
import json
from typing import List, Literal, Optional

# 3rd Party Modules
import urllib3

# Local Modules
from cofactr.schema import (
    OfferSchemaName,
    OrgSchemaName,
    ProductSchemaName,
    schema_to_offer,
    schema_to_org,
    schema_to_product,
)

Protocol = Literal["http", "https"]


drop_none_values = lambda d: {k: v for k, v in d.items() if v is not None}


def get_products(
    http, url, query, fields, before, after, limit, external, schema
):  # pylint: disable=too-many-arguments
    """Get products."""
    res = http.request(
        "GET",
        f"{url}/products",
        fields=drop_none_values(
            {
                "q": query,
                "fields": fields,
                "before": before,
                "after": after,
                "limit": limit,
                "external": external,
                "schema": schema,
            }
        ),
    )

    return json.loads(res.data.decode("utf-8"))


def get_orgs(
    http, url, query, before, after, limit, schema
):  # pylint: disable=too-many-arguments
    """Get orgs."""
    res = http.request(
        "GET",
        f"{url}/orgs",
        fields=drop_none_values(
            {
                "q": query,
                "before": before,
                "after": after,
                "limit": limit,
                "schema": schema,
            }
        ),
    )

    return json.loads(res.data.decode("utf-8"))


class GraphAPI:
    """A client-side representation of the Cofactr graph API."""

    PROTOCOL: Protocol = "https"
    HOST = "graph.cofactr.com"

    def __init__(
        self, protocol: Optional[Protocol] = PROTOCOL, host: Optional[str] = HOST
    ):

        self.url = f"{protocol}://{host}"
        self.http = urllib3.PoolManager()

    def check_health(self):
        """Check the operational status of the service."""

        res = self.http.request("GET", self.url)

        return json.loads(res.data.decode("utf-8"))

    def get_products(  # pylint: disable=too-many-arguments
        self,
        query: Optional[str] = None,
        fields: Optional[str] = None,
        before: Optional[str] = None,
        after: Optional[str] = None,
        limit: Optional[int] = None,
        external: Optional[bool] = True,
        schema: Optional[ProductSchemaName] = None,
    ):
        """Get products.

        Args:
            query: Search query.
            fields: Used to filter properties that the response should contain. A field can be a
                concrete property like "mpn" or an abstract group of properties like "assembly".
                Example: "id,aliases,labels,statements{spec,assembly},offers"
            before: Upper page boundry, expressed as a product ID.
            after: Lower page boundry, expressed as a product ID.
            limit: Restrict the results of the query to a particular number of documents.
            external: Whether to query external sources.
            schema: Response schema.
        """

        res = get_products(
            http=self.http,
            url=self.url,
            query=query,
            fields=fields,
            external=external,
            before=before,
            after=after,
            limit=limit,
            schema=schema.value,
        )

        Product = schema_to_product[schema]  # pylint: disable=invalid-name

        res["data"] = [Product(**data) for data in res["data"]]

        return res
    
    def get_parts_by_ids(
        self,
        ids: List[str],
        external: Optional[bool] = True,
        schema: ProductSchemaName = ProductSchemaName.FLAGSHIP,
    ):
        """Get a batch of parts.

        Note:
            Will evolve to use a batched requests. Where, for example, each request
            contains 50 part IDs.
        """
        with ThreadPoolExecutor() as executor:
            return dict(
                zip(
                    ids,
                    executor.map(
                        lambda cpid: self.get_product(id=cpid, external=external, schema=schema), ids
                    ),
                )
            )

    def get_orgs(  # pylint: disable=too-many-arguments
        self,
        query: Optional[str] = None,
        before: Optional[str] = None,
        after: Optional[str] = None,
        limit: Optional[int] = None,
        schema: Optional[str] = None,
    ):
        """Get organizations.

        Args:
            query: Search query.
            before: Upper page boundry, expressed as a product ID.
            after: Lower page boundry, expressed as a product ID.
            limit: Restrict the results of the query to a particular number of documents.
            schema: Response schema.
        """

        res = get_orgs(
            http=self.http,
            url=self.url,
            query=query,
            before=before,
            after=after,
            limit=limit,
            schema=schema.value,
        )

        Org = schema_to_org[schema]  # pylint: disable=invalid-name

        res["data"] = [Org(**data) for data in res["data"]]

        return res

    def get_product(
        self,
        id: str,
        fields: Optional[str] = None,
        external: Optional[bool] = True,
        schema: Optional[str] = None,
    ):
        """Get product.

        Args:
            fields: Used to filter properties that the response should contain. A field can be a
                concrete property like "mpn" or an abstract group of properties like "assembly".
                Example: "id,aliases,labels,statements{spec,assembly},offers"
            external: Whether to query external sources in order to update information for the
                given product.
            schema: Response schema.
        """

        res = json.loads(
            self.http.request(
                "GET",
                f"{self.url}/products/{id}",
                fields=drop_none_values(
                    {
                        "fields": fields,
                        "external": external,
                        "schema": schema.value,
                    }
                ),
            ).data.decode("utf-8")
        )

        Product = schema_to_product[schema]  # pylint: disable=invalid-name

        res["data"] = Product(**res["data"]) if (res and res.get("data")) else None

        return res

    def get_offers(
        self,
        product_id: str,
        fields: Optional[str] = None,
        external: Optional[bool] = True,
        schema: Optional[OfferSchemaName] = None,
    ):
        """Get product.

        Args:
            product_id: ID of the product to get offers for.
            fields: Used to filter properties that the response should contain.
            external: Whether to query external sources in order to update information.
            schema: Response schema.
        """

        res = json.loads(
            self.http.request(
                "GET",
                f"{self.url}/products/{product_id}/offers",
                fields=drop_none_values(
                    {
                        "fields": fields,
                        "external": external,
                        "schema": schema.value,
                    }
                ),
            ).data.decode("utf-8")
        )

        Offer = schema_to_offer[schema]  # pylint: disable=invalid-name

        res["data"] = [Offer(**data) for data in res["data"]]

        return res

    def get_org(self, id: str, schema: Optional[OrgSchemaName] = None):
        """Get organization."""

        res = json.loads(
            self.http.request(
                "GET", f"{self.url}/orgs/{id}", fields=drop_none_values({"schema": schema.value})
            ).data.decode("utf-8")
        )

        Org = schema_to_org[schema]  # pylint: disable=invalid-name

        res["data"] = Org(**res["data"]) if (res and res.get("data")) else None

        return res
